"""
test_server
----------------------------------

Tests for the Messenger class.
"""

import pytest
import aiohttp
import json
import hashlib
from sanic.utils import sanic_endpoint_test
from yarl import URL

from boomerang.server import Messenger
from boomerang.messages import Message
from boomerang.exceptions import BoomerangException, MessengerAPIException
from boomerang.events import MessageReceived
from boomerang.buttons import URLButton, PostbackButton


@pytest.fixture
def bot():
    '''Creates a new instance of the Messenger class.'''
    return Messenger('dummy_verify_token', 'dummy_page_token')


def test_run(bot, monkeypatch):
    '''Ensures that the run() method successfully executes.'''

    result = {}

    # To test this method, we monkeypatch the serve_multiple method of
    # the server, which will inform us when the server has successfully
    # started.
    def mock_serve(server_settings, workers):
        result['successful'] = True

    monkeypatch.setattr(bot._server, 'serve_multiple', mock_serve)

    # Run the server with two workers, in order to invoke serve_multiple()
    bot.run(processes=2)

    # Ensure the server started successfully
    assert result['successful']


def test_handle_api_error(bot):
    '''Tests the handle_api_error() method.'''
    response_json = {'error': {'message': 'Too many send requests to phone numbers',
                               'type': 'OAuthException',
                               'code': 4,
                               'error_subcode': 2018022,
                               'fbtrace_id': 'BLBaaaaaaa'}}
    exception_value = '4 (2018022): Too many send requests to phone numbers'

    with pytest.raises(MessengerAPIException) as exception:
        bot.handle_api_error(response_json)
        assert exception.value == exception_value


@pytest.mark.asyncio
async def test_post(bot, monkeypatch, event_loop):
    '''Tests the post() method.'''

    test_json = {'var1': 1, 'var2': 2}
    test_url = URL('http://www.google.com')

    # Create a ClientResponse object that will be returned as a result of the
    # mocked request.
    mock_response = aiohttp.client_reqrep.ClientResponse('post',
                                                         test_url)

    # Encode the test JSON and store it in the ClientResponse
    mock_response._content = bytes(json.dumps(test_json), encoding='ASCII')

    # Add empty headers to the ClientResponse
    mock_response.headers = aiohttp.CIMultiDictProxy(aiohttp.CIMultiDict([]))

    # A function which mocks aiohttp's ClientSession._request to return
    # the mock response.
    async def mock_request(method, url, *args, **kwargs):
        return mock_response

    monkeypatch.setattr(aiohttp.client.ClientSession, '_request', mock_request)

    # Make the post request
    async with aiohttp.ClientSession(loop=event_loop) as session:
        response = await bot.post(session, "{}")

    # Check that the response matches the desired JSON
    assert response == test_json


@pytest.mark.asyncio
async def test_send(bot, monkeypatch):
    '''Tests the send() function, checking that the desired
    message ID is returned.'''

    message = Message(text='dummy_text')
    response_json = {'recipient_id': 123,
                     'message_id': 'dummy_message_id'}

    # Monkeypatch the Messenger.post method to return the desired
    # JSON
    async def mock_post(session, json_string):
        return response_json

    monkeypatch.setattr(bot, 'post', mock_post)

    # Ensure the response's message ID is correctly handled
    assert await bot.send(123, message) == 'dummy_message_id'


@pytest.mark.asyncio
async def test_send_error(bot, monkeypatch):
    '''Tests the send() function, when the Send API returns an error.'''

    message = Message(text='dummy_text')
    response_json = {'error': {'message': 'Too many send requests to phone numbers',
                               'type': 'OAuthException',
                               'code': 4,
                               'error_subcode': 2018022,
                               'fbtrace_id': 'BLBaaaaaaa'}}

    # Monkeypatch the Messenger.post method to return the desired
    # JSON
    async def mock_post(session, json_string):
        return response_json

    monkeypatch.setattr(bot, 'post', mock_post)

    # Ensure the response is returned verbatim
    with pytest.raises(MessengerAPIException):
        await bot.send(123, message) == response_json


@pytest.mark.asyncio
async def test_send_action_error(bot, monkeypatch):
    '''Tests the send_action() function, when the Send API returns
    an error.'''

    response_json = {'error': {'message': 'Too many send requests to phone numbers',
                               'type': 'OAuthException',
                               'code': 4,
                               'error_subcode': 2018022,
                               'fbtrace_id': 'BLBaaaaaaa'}}

    # Monkeypatch the Messenger.post method to return the desired
    # JSON
    async def mock_post(session, json_string):
        return response_json

    monkeypatch.setattr(bot, 'post', mock_post)

    with pytest.raises(MessengerAPIException):
        await bot.send_action(123, 'typing_on')


@pytest.mark.asyncio
async def test_send_action_invalid(bot):
    '''Tests the send_action() function when called with an invalid
    action.'''

    with pytest.raises(BoomerangException):
        await bot.send_action(123, 'invalid_action')


@pytest.mark.asyncio
async def test_send_action(bot, monkeypatch):
    '''Tests the send_action() function, checking that the desired
    message ID is returned.'''

    response_json = {'recipient_id': 123,
                     'message_id': 'dummy_message_id'}

    # Monkeypatch the Messenger.post method to return the desired
    # JSON
    async def mock_post(session, json_string):
        return response_json

    monkeypatch.setattr(bot, 'post', mock_post)

    # Ensure the response's message ID is correctly handled
    assert await bot.send_action(123, 'typing_on') == 'dummy_message_id'


@pytest.mark.asyncio
async def test_acknowledge(bot, monkeypatch):
    '''Tests the acknowledge() function, ensuring it calls send_action
    with the appropriate arguments.'''
    result = {}

    # Create a MessageReceived event
    json_data = {'mid': 'mid.1482375186449:88d829fb30',
                 'seq': 426185,
                 'text': 'ping'}

    message_received_event = MessageReceived.from_json(123, 1234567890,
                                                       json_data)

    # Mock the send_action method to record all actions sent in the result
    # dictionary
    async def mock_send_action(recipient_id, action):
        if recipient_id == 123:
            result[action] = True

    monkeypatch.setattr(bot, 'send_action', mock_send_action)

    # Ensure that all required actions were sent
    await bot.acknowledge(message_received_event)
    assert result['mark_seen'] and result['typing_on']


@pytest.mark.asyncio
async def test_respond_to(bot, monkeypatch):
    result = {}

    # Create a MessageReceived event
    json_data = {'mid': 'mid.1482375186449:88d829fb30',
                 'seq': 426185,
                 'text': 'ping'}

    message_received_event = MessageReceived.from_json(123, 1234567890,
                                                       json_data)

    # Create a response Message
    response = Message(text='dummy_text')

    # Mock the send() function to record arguments it receives
    async def mock_send(recipient_id, message):
        result['recipient_id'] = recipient_id
        result['message'] = message

    monkeypatch.setattr(bot, 'send', mock_send)

    # Ensure respond_to calls send() with the appropriate arguments
    await bot.respond_to(message_received_event, response)
    assert result['recipient_id'] == 123
    assert result['message'] == response


@pytest.mark.asyncio
async def test_host_attachment(bot, monkeypatch):
    '''Tests the host_attachment() method by ensuring static hosting
    is correctly initialised and the correct MediaAttachment is returned.'''

    # Set a counter for the number of times static() is called.
    result = {'called': 0}

    # Generate the required URL
    relative_url = '/' + hashlib.sha256(bytes('image.png', 'utf-8')).hexdigest()

    # Set the base_url of the Messenger instance, which would normally be
    # set when run() is called.
    bot._base_url = 'http://base.com'

    # Mock the static() function to record arguments it receives
    def mock_static(url, filename):
        result['URL'] = url
        result['filename'] = filename

        # Increment the number of times static() is called
        # (it should only be once)
        result['called'] += 1

    monkeypatch.setattr(bot._server, 'static', mock_static)

    # Host a new attachment
    attachment = await bot.host_attachment('image', 'image.png')

    assert attachment.url == 'http://base.com' + relative_url
    assert result['URL'] == relative_url
    assert result['filename'] == 'image.png'

    # Re-host the same attachment, checking that it is not re-served.
    attachment = await bot.host_attachment('image', 'image.png')

    assert attachment.url == 'http://base.com' + relative_url
    assert result['URL'] == relative_url
    assert result['filename'] == 'image.png'
    assert result['called'] == 1


@pytest.mark.asyncio
async def test_host_attachment_no_base_url(bot):
    '''Tests that the host_attachment() method raises an error when
    base_url hasn't been set.'''
    bot._base_url = None
    with pytest.raises(BoomerangException):
        await bot.host_attachment('image', 'image.png')

@pytest.mark.asyncio
async def test_set_thread_settings(bot, monkeypatch):
    '''Tests the set_thread_settings method by ensuring it calls post()
    with the correct parameters.'''

    # The result dictionary holds a list of booleans
    result = {'success': []}

    # Mock the post() function to validate arguments it receives
    async def mock_post(session, request, api_endpoint='messages'):

        # This really ugly list contains the correct request JSON
        valid = [{'account_linking_url': 'https://cadelwatson.com',
                  'setting_type': 'account_linking'},
                 {'domain_action_type': 'add',
                  'setting_type': 'domain_whitelisting',
                  'whitelisted_domains': ['https://google.com']},
                 {'call_to_actions': [{'payload': 'payload'}],
                  'setting_type': 'call_to_actions',
                  'thread_state': 'new_thread'},
                 {'greeting': {'text': 'hi there'}, 'setting_type': 'greeting'},
                 {'call_to_actions': [{'title': 'Google',
                                       'type': 'web_url',
                                       'url': 'http://www.google.com'},
                                      {'payload': 'mmmmmm',
                                       'title': 'Hit me',
                                       'type': 'postback'}],
                  'setting_type': 'call_to_actions',
                  'thread_state': 'existing_thread'}]

        # Check that the request given to post() is valid
        if request in valid and api_endpoint == 'thread_settings':
            result['success'].append(True)
        else:
            result['success'].append(False)

        # Return a dummy response
        return {'result': 'success'}

    monkeypatch.setattr(bot, 'post', mock_post)

    # Create a thread settings request
    button_list = [URLButton('Google', 'http://www.google.com'),
                   PostbackButton('Hit me', 'mmmmmm')]

    await bot.set_thread_settings(menu_buttons=button_list,
                                  account_link_url='https://cadelwatson.com',
                                  whitelisted_domains=['https://google.com'],
                                  get_started_payload='payload',
                                  greeting_text='hi there')

    # Ensure that all post() requests were valid
    assert all(result['success'])


def test_valid_register(bot):
    '''Tests valid registration request by the Messenger Platform.'''
    valid_request = {'hub.mode': 'subscribe',
                     'hub.verify_token': 'dummy_verify_token',
                     'hub.challenge': 'dummy_challenge'}

    response = sanic_endpoint_test(bot._server,
                                   gather_request=False,
                                   uri='/webhook',
                                   params=valid_request)

    assert response.text == valid_request['hub.challenge']
    assert response.status == 200


def test_invalid_register(bot):
    '''Tests the response when verify_tokens do not match.'''
    invalid_request = {'hub.mode': 'subscribe',
                       'hub.verify_token': 'wrong_verify_token',
                       'hub.challenge': 'dummy_challenge'}

    response = sanic_endpoint_test(bot._server,
                                   gather_request=False,
                                   uri='/webhook',
                                   params=invalid_request)

    # Ensure the server doesn't provide the correct challenge anyway
    assert response.text != invalid_request['hub.challenge']

    # Check that the server replied that access was forbidden
    assert response.status == 403


def message_handled_ok(server, message_type, message_content):
    '''Makes a POST request to the server, simulating the given message being
    sent, and ensure the server responds with a 200 OK status.

    Args:
        server: A Sanic server to test against.
        message_type: A string of the type of message, given as the Messenger
                      API string for the type.
        message_content: A dict representing the JSON message to be tested.

    Returns:
        True if a 200 OK response is received, otherwise False.

    '''

    message = {'entry': [{'id': '233231370449158',
                          'messaging': [{message_type: message_content,
                                         'recipient': {'id': '233231370449158'},
                                         'sender': {'id': '1297601746979209'},
                                         'timestamp': 1482375186449}],
                          'time': 1482375186497}],
               'object': 'page'}

    message_json_string = json.dumps(message)

    request, response = sanic_endpoint_test(server,
                                            method='post',
                                            uri='/webhook',
                                            data=message_json_string)

    return response.status == 200


def test_message_received(bot):
    '''Ensure the server responds with a 200 status to a message received event
    sent by the API.'''
    message = {'mid': 'mid.1482375186449:88d829fb30',
               'seq': 426185,
               'text': 'ping'}
    assert message_handled_ok(bot._server, 'message', message)


def test_message_delivered(bot):
    '''Ensure the server responds with a 200 status to a message delivered event
    sent by the API.'''
    message = {'mids': ['mid.1458668856218:ed81099e15d3f4f233',
                        'mid.1458668856218:ed81099e15d3f4f234'],
               'watermark': 1111111111,
               'seq': 30}

    assert message_handled_ok(bot._server, 'delivery', message)


def test_message_read(bot):
    '''Ensure the server responds with a 200 status to a message read event
    sent by the API.'''
    message = {'watermark': 1111111111,
               'seq': 30}

    assert message_handled_ok(bot._server, 'read', message)


def test_postback(bot):
    '''Ensure the server responds with a 200 status to a postback event
    sent by the API.'''
    message = {'payload': 'dummy_payload'}

    assert message_handled_ok(bot._server, 'postback', message)


def test_postback_with_referral(bot):
    '''Ensure the server responds with a 200 status to a postback event
    sent by the API, when the postback contains a referral.'''
    message = {'payload': 'dummy_payload',
               'referral': {'ref': 'dummy_referral_data',
                            'source': 'SHORTLINK',
                            'type': 'OPEN_THREAD'}}

    assert message_handled_ok(bot._server, 'postback', message)


def test_referral(bot):
    '''Ensure the server responds with a 200 status to a referral event
    sent by the API.'''
    message = {'ref': 'dummy_referral_data',
               'source': 'SHORTLINK',
               'type': 'OPEN_THREAD'}

    assert message_handled_ok(bot._server, 'referral', message)


def test_optin(bot):
    '''Ensure the server responds with a 200 status to a plugin opt in event
    sent by the API.'''
    message = {'ref': 'dummy_data'}

    assert message_handled_ok(bot._server, 'optin', message)


def test_account_linking(bot):
    '''Ensure the server responds with a 200 status to an account linking event
    sent by the API.'''
    message = {'status': 'linked',
               'authorization_code': 'dummy_authorization'}

    assert message_handled_ok(bot._server, 'account_linking', message)



import abc

from .exceptions import BoomerangException

class Button(metaclass=abc.ABCMeta):
    '''The abstract base class for all button types.

    Attributes:
        button_type: The type of the button.
        text: The text to display on the button. Due to the Send API's
              limitation, it cannot be more than 20 characters long.

    '''
    def __init__(self, button_type, text):
        self.button_type = button_type

        if len(text) > 20:
            error = 'Text on buttons may not be more than 20 characters long'
            raise BoomerangException(error)
        else:
            self.text = text


class URLButton(Button):
    '''A class representing a URL button.

    Attributes:
        text: The text to display on the button.
        url: The url to open when the button is pressed.

    '''
    def __init__(self, text, url):
        super().__init__('web_url', text)

        self.url = url

    def to_json(self):
        '''Converts the button to JSON.

        Returns:
            A dictionary holding the JSON representation of the button.

        '''
        return {'type': self.button_type,
                'title': self.text,
                'url': self.url}


class PostbackButton(Button):
    '''A class representing a postback button.

    Attributes:
        text: The text to display on the button.
        payload: The string to send to the webhook as a postback, when the
                 button is clicked.

    '''
    def __init__(self, text, payload):
        super().__init__('postback', text)

        self.payload = payload

    def to_json(self):
        '''Converts the button to JSON.

        Returns:
            A dictionary holding the JSON representation of the button.

        '''
        return {'type': self.button_type,
                'title': self.text,
                'payload': self.payload}


class ShareButton(Button):
    '''A class representing a share button, which can only be present inside
    a generic template.

    '''
    def __init__(self):
        self.button_type = 'element_share'

    def to_json(self):
        '''Converts the button to JSON.

        Returns:
            A dictionary holding the JSON representation of the button.

        '''
        return {'type': self.button_type}


class CallButton(Button):
    '''A class representing a call button, which triggers a phone call.

    Attributes:
        text: The text to display on the button.
        phone_number: A string representing a phone number
                      (including area code).

    '''
    def __init__(self, text, phone_number):
        super().__init__('phone_number', text)

        self.phone_number = phone_number

    def to_json(self):
        '''Converts the button to JSON.

        Returns:
            A dictionary holding the JSON representation of the button.

        '''
        return {'type': self.button_type,
                'title': self.text,
                'payload': self.phone_number}


class QuickReply:
    '''A class representing quick-reply buttons.

    Attributes:
        button_type: The type of the button, either 'text' or 'location'.
        text: The text to display on the button. Due to the Send API's
              limitation, it cannot be more than 20 characters long. Only
              available when button_type is 'text'.
        payload: The payload string that will be returned to the webhook when
                 the button is pressed. Only available when button_type is
                 'text'.
        image_url: The URL of an image to use on the button. Only available
                   when button_type is 'text'.

    '''
    def __init__(self, button_type, text=None, payload=None, image_url=None):

        # Ensure that only a valid button type is chosen.
        if button_type not in ['text', 'location']:
            error = "QuickReply button_type must be 'text' or 'location'"
            raise BoomerangException(error)

        self.button_type = button_type

        # Initialise a text button.
        if button_type == 'text':
            if text is None or payload is None:
                error = "QuickReply text button must contain text and payload"
                raise BoomerangException(error)

            self.text = text
            self.payload = payload
            self.image_url = image_url

        # Initialise a location button, which can't take other parameters.
        else:
            if any(attr is not None for attr in [text, payload, image_url]):
                error = ('QuickReply location button cannot contain other'
                         'attributes')
                raise BoomerangException(error)

    def to_json(self):
        '''Converts the button to JSON.

        Returns:
            A dictionary holding the JSON representation of the button.

        '''
        json = {'content_type': self.button_type}

        if self.button_type == 'text':
            json['title'] = self.text
            json['payload'] = self.payload

            if self.image_url is not None:
                json['image_url'] = self.image_url

        return json



from .messages import LocationAttachment, MediaAttachment


class MessageReceived():
    '''A class holding Message Received events.

    Attributes:
        user_id: An integer Messenger ID of the user who triggered the
                 event.
        timestamp: An integer timestamp (in milliseconds since epoch) of
                   the event.
        text: A string representing the text of the message.
        attachments: A list of Attachment objects.
        quick_reply: A QuickReplyPayload holding a payload set by
                     the bot for quick replies.
        message_id: An integer Messenger ID of the message.
        sequence_position: An integer position of the event in the
                           conversation sequence.

    '''

    def __init__(self, user_id, timestamp, text, attachments,
                 quick_reply, message_id, sequence_position):
        self.user_id = user_id
        self.timestamp = timestamp
        self.text = text
        self.attachments = attachments
        self.quick_reply = quick_reply
        self.message_id = message_id
        self.sequence_position = sequence_position

    @classmethod
    def from_json(cls, user_id, timestamp, json):
        '''Builds a MessageReceived object from JSON provided by the API.

        Args:
            user_id: An integer Messenger ID of the user who triggered the
                    event.
            timestamp: An integer timestamp (in milliseconds since epoch) of
                    the event.
            json: A dict containing the JSON representation that is converted
                  into the MessageReceived object.

        Returns:
            A MessageReceived object.

        '''
        # Convert any quick reply to a Quick Reply object
        if 'quick_reply' in json:
            quick_reply = QuickReplyPayload.from_json(json['quick_reply'])
        else:
            quick_reply = None

        # Convert all attachments to Attachment objects
        attachments = []
        for attachment in json.get('attachments', []):
            if attachment['type'] == 'location':
                attachments.append(LocationAttachment.from_json(attachment))
            else:
                attachments.append(MediaAttachment.from_json(attachment))

        # Get text of message if available
        text = json.get('text', '')

        return cls(user_id, timestamp, text=text,
                   attachments=attachments, quick_reply=quick_reply,
                   message_id=json['mid'], sequence_position=json['seq'])


class QuickReplyPayload:
    '''A class holding a developer-specified quick reply payload.

    Attributes:
        payload: The string payload of the quick reply returned by the API.

    '''

    def __init__(self, payload):
        self.payload = payload

    @classmethod
    def from_json(cls, json):
        '''Builds a QuickReplyPayload object from JSON provided by the API.

        Args:
            json: A dict containing the JSON representation that is converted
                  into the QuickAttachment object.

        Returns:
            A QuickReplyPayload object.

        '''
        return cls(json['payload'])


class MessageDelivered():
    '''A class holding Message Delivered events.

    Attributes:
        user_id: An integer Messenger ID of the user who triggered the
                 event.
        timestamp: An integer timestamp (in milliseconds since epoch) of
                   the event.
        watermark: An integer timestamp. All messages with a timestamp before
                   this watermark are guaranteed to have been delivered.
        message_ids: A list of string Messenger IDs of the delivered
                     messages.
        sequence_position: An integer position of the event in the
                           conversation sequence.

    '''

    def __init__(self, user_id, timestamp, watermark, message_ids,
                 sequence_position):
        self.user_id = user_id
        self.timestamp = timestamp
        self.watermark = watermark
        self.message_ids = message_ids
        self.sequence_position = sequence_position

    @classmethod
    def from_json(cls, user_id, timestamp, json):
        '''Builds a MessageDelivered object from JSON provided by the API.

        Args:
            user_id: An integer Messenger ID of the user who triggered the
                    event.
            timestamp: An integer timestamp (in milliseconds since epoch) of
                    the event.
            json: A dict containing the JSON representation that is converted
                  into the MessageDelivered object.

        Returns:
            A MessageDelivered object.

        '''

        # If message IDs aren't present, store an empty list.
        message_ids = json.get('mids', [])

        return cls(user_id, timestamp, json['watermark'], message_ids,
                   json['seq'])


class MessageRead():
    '''A class holding Message Delivered events.

    Attributes:
        user_id: An integer Messenger ID of the user who triggered the
                 event.
        timestamp: An integer timestamp (in milliseconds since epoch) of
                   the event.
        watermark: An integer timestamp. All messages with a timestamp before
                   this watermark are guaranteed to have been read.
        sequence_position: An integer position of the event in the
                           conversation sequence.

    '''

    def __init__(self, user_id, timestamp, watermark, sequence_position):
        self.user_id = user_id
        self.timestamp = timestamp
        self.watermark = watermark
        self.sequence_position = sequence_position

    @classmethod
    def from_json(cls, user_id, timestamp, json):
        '''Builds a MessageRead object from JSON provided by the API.

        Args:
            user_id: An integer Messenger ID of the user who triggered the
                    event.
            timestamp: An integer timestamp (in milliseconds since epoch) of
                    the event.
            json: A dict containing the JSON representation that is converted
                  into the MessageRead object.

        Returns:
            A MessageRead object.

        '''

        return cls(user_id, timestamp, json['watermark'], json['seq'])


class Postback():
    '''A class holding Postback events.

    Attributes:
        user_id: An integer Messenger ID of the user who triggered the
                 event.
        timestamp: An integer timestamp (in milliseconds since epoch) of
                   the event.
        payload: A string defined by the bot when creating the postback button.
        referral: An optional Referral object.

    '''

    def __init__(self, user_id, timestamp, payload, referral=None):
        self.user_id = user_id
        self.timestamp = timestamp
        self.payload = payload
        self.referral = referral

    @classmethod
    def from_json(cls, user_id, timestamp, json):
        '''Builds a Postback object from JSON provided by the API.

        Args:
            user_id: An integer Messenger ID of the user who triggered the
                    event.
            timestamp: An integer timestamp (in milliseconds since epoch) of
                    the event.
            json: A dict containing the JSON representation that is converted
                  into the MessageRead object.

        Returns:
            A Postback object.

        '''

        # Convert a referral to a Referral object if present
        if 'referral' in json:
            referral = Referral.from_json(user_id, timestamp, json['referral'])
        else:
            referral = None

        return cls(user_id, timestamp, json['payload'], referral)


class Referral():
    '''A class holding Referral events.

    Attributes:
        user_id: An integer Messenger ID of the user who triggered the
                 event.
        timestamp: An integer timestamp (in milliseconds since epoch) of
                   the event.
        data: A string passed by the referral link.

    '''

    def __init__(self, user_id, timestamp, data):
        self.user_id = user_id
        self.timestamp = timestamp
        self.data = data

    @classmethod
    def from_json(cls, user_id, timestamp, json):
        '''Builds a Referral object from JSON provided by the API.

        Args:
            user_id: An integer Messenger ID of the user who triggered the
                    event.
            timestamp: An integer timestamp (in milliseconds since epoch) of
                    the event.
            json: A dict containing the JSON representation that is converted
                  into the Referral object.

        Returns:
            A Referral object.

        '''

        return cls(user_id, timestamp, json['ref'])


class OptIn():
    '''A class holding OptIn events.

    Attributes:
        user_id: An integer Messenger ID of the user who triggered the
                 event.
        timestamp: An integer timestamp (in milliseconds since epoch) of
                   the event.
        data: A string passed by the 'Send to Messenger' plugin.

    '''

    def __init__(self, user_id, timestamp, data):
        self.user_id = user_id
        self.timestamp = timestamp
        self.data = data

    @classmethod
    def from_json(cls, user_id, timestamp, json):
        '''Builds an OptIn object from JSON provided by the API.

        Args:
            user_id: An integer Messenger ID of the user who triggered the
                    event.
            timestamp: An integer timestamp (in milliseconds since epoch) of
                    the event.
            json: A dict containing the JSON representation that is converted
                  into the OptIn object.

        Returns:
            An OptIn object.

        '''

        return cls(user_id, timestamp, json['ref'])


class AccountLink():
    '''A class holding AccountLink events.

    Attributes:
        user_id: An integer Messenger ID of the user who triggered the
                 event.
        timestamp: An integer timestamp (in milliseconds since epoch) of
                   the event.
        status: A string representing the account action. Can be 'linked' or
                'unlinked'.
        authorization_code: An authorization code provided when status is
                            'linked'. The code is defined by the user during
                            the account linking process.

    '''

    def __init__(self, user_id, timestamp, status, authorization_code=None):
        self.user_id = user_id
        self.timestamp = timestamp
        self.status = status
        self.authorization_code = authorization_code

    @classmethod
    def from_json(cls, user_id, timestamp, json):
        '''Builds an AccountLink object from JSON provided by the API.

        Args:
            user_id: An integer Messenger ID of the user who triggered the
                    event.
            timestamp: An integer timestamp (in milliseconds since epoch) of
                    the event.
            json: A dict containing the JSON representation that is converted
                  into the AccountLink object.

        Returns:
            An AccountLink object.

        '''

        # If the account is being unlinked, no authorization code is provided.
        if json['status'] == 'unlinked':
            return cls(user_id, timestamp, json['status'])
        else:
            return cls(user_id, timestamp, json['status'],
                       json['authorization_code'])



"""
test_events
----------------------------------

Tests for classes in the events module.
"""

from boomerang import events


def test_message_received():
    '''Tests the to_json and from_json functionality of the basic Message class.

    '''
    json_data = {'mid': 'mid.1482375186449:88d829fb30',
                 'seq': 426185,
                 'text': 'ping'}
    user_id = 12345
    timestamp = 1234567890

    message_object = events.MessageReceived.from_json(user_id, timestamp,
                                                      json_data)

    assert message_object.user_id == user_id
    assert message_object.timestamp == timestamp
    assert message_object.message_id == json_data['mid']
    assert message_object.sequence_position == json_data['seq']
    assert message_object.text == json_data['text']


def test_message_received_with_attachments():
    '''Tests the to_json and from_json functionality of the basic Message class,
    with attachments included.

    '''
    image_attachment = {'payload': {'url': 'http://www.google.com'},
                        'type': 'image'}
    location_attachment = {'title': "Someones's Location",
                           'type': 'location',
                           'url': 'dummy_url',
                           'payload': {'coordinates': {'long': 38.8976763,
                                                       'lat': -77.0387185}}}
    json_data = {'attachments': [image_attachment, location_attachment],
                 'mid': 'mid.1482375360960:63e41d7f60',
                 'seq': 426187}
    user_id = 12345
    timestamp = 1234567890

    message_object = events.MessageReceived.from_json(user_id, timestamp,
                                                      json_data)

    assert len(message_object.attachments) == 2
    assert isinstance(message_object.attachments[0],
                      events.MediaAttachment)
    assert isinstance(message_object.attachments[1],
                      events.LocationAttachment)


def test_message_received_with_quickreply():
    '''Tests the to_json and from_json functionality of the basic Message class,
    with a quick reply included.

    '''
    json_data = {'mid': 'mid.1457764197618:41d102a3e1ae206a38',
                 'seq': 12345,
                 'text': 'dummy_text',
                 'quick_reply': {'payload': 'dummy_payload'}}

    user_id = 12345
    timestamp = 1234567890

    message_object = events.MessageReceived.from_json(user_id, timestamp,
                                                      json_data)

    assert isinstance(message_object.quick_reply, events.QuickReplyPayload)
    assert message_object.quick_reply.payload == 'dummy_payload'


def test_message_delivered():
    '''Tests the to_json functionality of the MessageDelivered class.

    '''
    json_data = {'mids': ['mid.1458668856218:ed81099e15d3f4f233',
                          'mid.1458668856218:ed81099e15d3f4f234'],
                 'watermark': 1111111111,
                 'seq': 30}

    user_id = 12345
    timestamp = 1234567890

    message_object = events.MessageDelivered.from_json(user_id, timestamp,
                                                       json_data)

    assert message_object.user_id == user_id
    assert message_object.timestamp == timestamp
    assert message_object.watermark == json_data['watermark']
    assert message_object.sequence_position == json_data['seq']

    assert isinstance(message_object.message_ids, list)
    assert message_object.message_ids[0] == 'mid.1458668856218:ed81099e15d3f4f233'


def test_message_read():
    '''Tests the to_json functionality of the MessageRead class.

    '''
    json_data = {'watermark': 1111111111,
                 'seq': 30}

    user_id = 12345
    timestamp = 1234567890

    message_object = events.MessageRead.from_json(user_id, timestamp,
                                                  json_data)

    assert message_object.user_id == user_id
    assert message_object.timestamp == timestamp
    assert message_object.watermark == json_data['watermark']
    assert message_object.sequence_position == json_data['seq']


def test_postback():
    '''Tests the to_json functionality of the Postback class.

    '''
    json_data = {'payload': 'dummy_payload'}

    user_id = 12345
    timestamp = 1234567890

    message_object = events.Postback.from_json(user_id, timestamp, json_data)

    assert message_object.user_id == user_id
    assert message_object.timestamp == timestamp
    assert message_object.payload == 'dummy_payload'
    assert message_object.referral is None


def test_postback_with_referral():
    '''Tests the to_json functionality of the Postback class, when a referral is
    included.

    '''
    json_data = {'payload': 'dummy_payload',
                 'referral': {'ref': 'dummy_referral_data',
                              'source': 'SHORTLINK',
                              'type': 'OPEN_THREAD'}}

    user_id = 12345
    timestamp = 1234567890

    message_object = events.Postback.from_json(user_id, timestamp, json_data)

    assert message_object.user_id == user_id
    assert message_object.timestamp == timestamp
    assert message_object.payload == 'dummy_payload'

    assert isinstance(message_object.referral, events.Referral)
    assert message_object.referral.data == 'dummy_referral_data'


def test_referral():
    '''Tests the to_json functionality of the Referral class.

    '''
    json_data = {'ref': 'dummy_referral_data',
                 'source': 'SHORTLINK',
                 'type': 'OPEN_THREAD'}

    user_id = 12345
    timestamp = 1234567890

    message_object = events.Referral.from_json(user_id, timestamp, json_data)

    assert message_object.user_id == user_id
    assert message_object.timestamp == timestamp
    assert message_object.data == 'dummy_referral_data'


def test_optin():
    '''Tests the to_json functionality of the OptIn class.

    '''
    json_data = {'ref': 'dummy_data'}

    user_id = 12345
    timestamp = 1234567890

    message_object = events.OptIn.from_json(user_id, timestamp, json_data)

    assert message_object.user_id == user_id
    assert message_object.timestamp == timestamp
    assert message_object.data == 'dummy_data'


def test_account_unlinked():
    '''Tests the to_json functionality of the AccountLink class, when
    the account is being unlinked.

    '''

    json_data = {'status': 'unlinked'}

    user_id = 12345
    timestamp = 1234567890

    message_object = events.AccountLink.from_json(user_id, timestamp,
                                                  json_data)

    assert message_object.user_id == user_id
    assert message_object.timestamp == timestamp
    assert message_object.status == 'unlinked'
    assert message_object.authorization_code is None


def test_account_linked():
    '''Tests the to_json functionality of the AccountLink class, when
    the account is being linked.

    '''

    json_data = {'status': 'linked',
                 'authorization_code': 'dummy_authorization'}

    user_id = 12345
    timestamp = 1234567890

    message_object = events.AccountLink.from_json(user_id, timestamp,
                                                  json_data)

    assert message_object.user_id == user_id
    assert message_object.timestamp == timestamp
    assert message_object.status == 'linked'
    assert message_object.authorization_code == 'dummy_authorization'


def test_message_delivered_with_no_message_ids():
    '''Tests the to_json functionality of the MessageDelivered class,
    when the array of message IDs is missing.

    '''
    json_data = {'watermark': 1111111111,
                 'seq': 30}

    user_id = 12345
    timestamp = 1234567890

    message_object = events.MessageDelivered.from_json(user_id, timestamp,
                                                       json_data)

    assert isinstance(message_object.message_ids, list)
    assert len(message_object.message_ids) == 0


def test_attachments():
    '''Tests the to_json and from_json functionality of the attachment classes.

    '''
    media_json_data = {'payload': {'url': 'http://www.google.com'},
                       'type': 'image'}

    attachment_object = events.MediaAttachment.from_json(media_json_data)
    assert attachment_object.media_type == 'image'
    assert attachment_object.url == 'http://www.google.com'

    location_json_data = {'title': "Someones's Location",
                          'type': 'location',
                          'url': 'dummy_url',
                          'payload': {'coordinates': {'long': 38.8976763,
                                                      'lat': -77.0387185}}}

    location_object = events.LocationAttachment.from_json(location_json_data)
    assert location_object.latitude == -77.0387185
    assert location_object.longitude == 38.8976763


def test_quickreply():
    '''Tests the from_json functionality of the QuickReplyPayload class.

    '''
    json_data = {'payload': 'dummy_payload'}

    quickreply_object = events.QuickReplyPayload.from_json(json_data)
    assert quickreply_object.payload == 'dummy_payload'



import uvloop
import json
import aiohttp
import hashlib

from sanic import Sanic
import sanic.response as response

from . import events, messages
from .exceptions import BoomerangException, MessengerAPIException


class Messenger:
    '''The base class that contains the Facebook Messenger bot, and handles
    webhooks and sending.

    The application should subclass this class and override the webhook handler
    methods as required.

    Attributes:
       verify_token: The string Messenger Platform verify token.
       page_token: The string Messenger Platform page access token.

    '''
    def __init__(self, verify_token, page_token):
        self._verify_token = verify_token
        self._page_token = page_token

        self._event_loop = uvloop.new_event_loop()
        self._server = Sanic(__name__)

        # A dictionary which holds currently-hosted attachments, where
        # the filename is the key and the relative url is the value.
        self._attachments = {}

        # Create a handler for the webhook which delegates to different
        # functions depending on the HTTP method used. GET requests are used
        # by Messenger to validate the bot, while POST requests are used to
        # send the bot events.
        @self._server.route('/webhook', methods=['GET', 'POST'])
        async def webhook(request):
            if request.method == 'GET':
                server_response = self.register(request)
                return server_response

            elif request.method == 'POST':
                server_response = await self.handle_webhook(request)
                return server_response

    def run(self, hostname='127.0.0.1', base_url=None, port=8000, debug=False,
            processes=1):
        '''Runs the bot using the given server settings.

        Args:
            hostname: A string representing the hostname on which to run the
                      server.
            base_url: The public url of the bot server, used for tasks such
                      as attachment hosting. If None, these features will not
                      be available.
            port: The integer port on which to run the server.
            debug: A boolean enabling debug logging during operation.
            processes: An integer number of processes to run the server on.

        Returns:
            None

        '''
        self._base_url = base_url
        self._server.run(loop=self._event_loop,
                         host=hostname,
                         port=port,
                         debug=debug,
                         workers=processes)

    def register(self, request):
        '''Handles registration of the Messenger Platform using the webhook system.

        Args:
            request: A request object passed by the Sanic server.

        Returns:
            A Sanic text response. If registration was successful, the provided
            challenge string is returned with a 200 OK status. However, if the
            verification token does not match the class' token, a 403 FORBIDDEN
            request is returned.

        '''
        request_type = request.args['hub.mode'][0]
        request_verify_token = request.args['hub.verify_token'][0]
        request_challenge = request.args['hub.challenge'][0]

        is_verified = (request_verify_token == self._verify_token)

        if request_type == 'subscribe':
            if is_verified:
                return response.text(request_challenge, status=200)
            else:
                return response.text('Verification token did not match server',
                                     status=403)

    async def handle_webhook(self, request):
        '''Handles all POST requests made to the /webhook endpoint by the Messenger
        Platform.

        The request is formatted and delegated to the relevant
        user-implementable event functions.

        Args:
            request: A request object passed by the Sanic server.

        Returns:
            A Sanic text response. If the request was successfully handled,
            the response is 200 OK.

        '''
        # The API provides a list of events inside of the 'entry' list
        for event in request.json['entry']:

            # Inside the event is a list of messages, under 'messaging'
            for message in event['messaging']:

                # Isolate the User ID and timestamp, which are both present
                # regardless of message type
                user_id = int(message['sender']['id'])
                timestamp = message['timestamp']

                # Each message type has an identifier, which Facebook uses to
                # signify the type; a class, for holding the message; and a
                # handler function which is implemented by the user.
                message_types = [('message', events.MessageReceived,
                                  self.message_received),
                                 ('delivery', events.MessageDelivered,
                                  self.message_delivered),
                                 ('read', events.MessageRead,
                                  self.message_read),
                                 ('postback', events.Postback,
                                  self.postback),
                                 ('referral', events.Referral,
                                  self.referral),
                                 ('optin', events.OptIn,
                                  self.opt_in),
                                 ('account_linking', events.AccountLink,
                                  self.account_link)]

                # Loop through the message types. If one is found, create the
                # relevant message object and handle it. Then break from the
                # loop as there can only be one type present at a time.
                for message_type in message_types:
                    identifier, message_class, handler = message_type

                    if identifier in message:
                        # Don't respond to echo events
                        if 'is_echo' not in message[identifier]:
                            event = message_class.from_json(user_id,
                                                            timestamp,
                                                            message[identifier])

                            await handler(event)
                            break

        return response.text('Success', status=200)

    def handle_api_error(self, error_json):
        '''Wraps the error JSON returned by Facebook in a
        MessengerAPIException and raises it.

        Args:
            error_json: A dictionary containing the JSON representation
                        of an error returned by the API.

        Returns:
            None

        Raises:
            MessengerAPIException: Wraps the API error.

        '''
        error_payload = error_json['error']
        error_message = '{0} ({1}): {2}'.format(error_payload['code'],
                                                error_payload['error_subcode'],
                                                error_payload['message'])

        raise MessengerAPIException(error_message)

    async def post(self, session, data, api_endpoint='messages'):
        '''Makes a POST request to the Send API.

        Args:
            session: The aiohttp session to use.
            data: A JSON dictionary that will be the body of the POST request.
            api_endpoint: The endpoint in the Facebook Graph API to post
                          to. The default is 'messages' for the Send API,
                          but it may be changed to allow for thread settings
                          etc.

        Returns:
            An aiohttp response in JSON format.
        '''
        url = 'https://graph.facebook.com/v2.6/me/' \
              + api_endpoint \
              + '?access_token=' \
              + self._page_token
        headers = {'content-type': 'application/json'}

        # Create a JSON string
        json_data = json.dumps(data)

        async with session.post(url, data=json_data, headers=headers) as response:
            return await response.json()

    async def send(self, recipient_id, message):
        '''Sends a message to the given recipient using the Send API.

        Args:
            recipient_id: The integer ID of the user to send the message to.
            message: A Message object containing the message to send.

        Returns:
            The message ID returned by the Send API.

        '''
        # Create the JSON for the POST request.
        json_message = {'recipient': {'id': recipient_id},
                        'message': message.to_json()}

        async with aiohttp.ClientSession(loop=self._event_loop) as session:
            response = await self.post(session, json_message)

        if 'message_id' in response:
            return response['message_id']
        else:
            self.handle_api_error(response)

    async def send_action(self, recipient_id, action):
        '''Sends an action to the given recipient using the Send API.

        Args:
            recipient_id: The integer ID of the user to send the message to.
            action: One of 'typing_on', 'typing_off', or 'mark_seen'.

        Returns:
            The message ID returned by the Send API.

        '''

        if action not in ['typing_on', 'typing_off', 'mark_seen']:
            error = ("Action must be one of 'typing_on', 'typing_off',"
                     "or 'mark_seen', not {0}".format(action))
            raise BoomerangException(error)

        # Create the JSON for the POST request.
        json_message = {'recipient': {'id': recipient_id},
                        'sender_action': action}

        async with aiohttp.ClientSession(loop=self._event_loop) as session:
            response = await self.post(session, json_message)

        if 'message_id' in response:
            return response['message_id']
        else:
            self.handle_api_error(response)

    async def acknowledge(self, message):
        '''Use message actions to acknowledge to the user that sent the given
        message that the message has been received and is being worked on.

        This function sends both the 'mark_seen' and 'typing_on' events to the
        user. Sending a new message to the user will automatically disable the
        typing notification.

        Args:
            message: A MessageReceived object to acknowledge the receipt of.

        '''
        await self.send_action(message.user_id, 'mark_seen')
        await self.send_action(message.user_id, 'typing_on')

    async def respond_to(self, message_received, response):
        '''Send a new Message to the user who triggered the given MessageReceived
        event.

        Args:
            message_received: A MessageReceived object to respond to.
            response: A Message object to send to the user who triggered
                      message_recieved.

        Returns:
            The message ID returned by the Send API.

        '''
        return await self.send(message_received.user_id, response)

    async def host_attachment(self, media_type, filename):
        '''Hosts the given file on the bot server, returning a MediaAttachment
        object which can be sent to the Messenger API, which will access the
        attachment from the current server.

        Args:
            media_type: The type of the attachment. Can be one of 'image',
                        'audio', 'video' or 'file'.
            filename: The filename of the attachment.

        Returns:
            A MediaAttachment object representing the hosted attachment.
        '''

        # If the base_url isn't set or is None, attachments cannot be hosted
        if self._base_url is None:
            error = 'base_url must be set for attachment hosting'
            raise BoomerangException(error)

        if filename not in self._attachments:
            # The SHA-256 hash of the filename is used as an identifier,
            # which is served as the URL of the attachment. This has two
            # benefits: firstly, it guarantees a unique identifier for
            # different attachments; and secondly, it obfuscates the file
            # structure of the hosting machine for better security.
            identifier = hashlib.sha256(bytes(filename,
                                              'utf-8')).hexdigest()
            self._attachments[filename] = '/' + identifier

            # Serve the attachment using the internal server
            self._server.static(self._attachments[filename], filename)

        # Create and return the MediaAttachment object
        relative_url = self._attachments[filename]
        full_url = self._base_url + relative_url
        return messages.MediaAttachment(media_type, full_url)

    async def set_thread_settings(self, account_link_url=None,
                                  whitelisted_domains=None,
                                  get_started_payload=None,
                                  greeting_text=None,
                                  menu_buttons=None):
        '''Calls the Messenger API to set the thread settings with the given
        parameters:

        Args:
            account_link_url: The URL to use for the account linking process.
            whitelisted_domains: A list of Domain URLs to whitelist for
                                extensions.
            get_started_payload: A string payload which will be sent to the
                                 webhook when the 'Get Started' button is
                                 pressed.
            greeting_text: A string which will be displayed to new users.
            menu_buttons: A list of buttons (either URLButton or
                          PostbackButton) to add to a persistent
                          menu on the application

        Returns:
            None

        '''
        # A queue of JSON requests which will be sent one-by-one to the API.
        request_queue = []

        if account_link_url is not None:
            request_queue.append({'setting_type': 'account_linking',
                                  'account_linking_url': account_link_url})

        if whitelisted_domains is not None:
            request_queue.append({'setting_type': 'domain_whitelisting',
                                  'domain_action_type': 'add',
                                  'whitelisted_domains': whitelisted_domains})

        if get_started_payload is not None:
            request_queue.append({'setting_type': 'call_to_actions',
                                  'thread_state': 'new_thread',
                                  'call_to_actions': [{'payload':
                                                       get_started_payload}]})

        if greeting_text is not None:
            request_queue.append({'setting_type': 'greeting',
                                  'greeting': {'text': greeting_text}})

        if menu_buttons is not None:
            request_queue.append({'setting_type': 'call_to_actions',
                                  'thread_state': 'existing_thread',
                                  'call_to_actions': [button.to_json() for
                                                      button in menu_buttons]})

        # Make each request in the queue
        async with aiohttp.ClientSession(loop=self._event_loop) as session:
            for request in request_queue:
                response = await self.post(session, request,
                                           api_endpoint='thread_settings')

                # Handle errors returned by the API
                if 'result' not in response:
                    self.handle_api_error(response)

    async def message_received(self, message):
        '''Handles all 'message received' events sent to the bot.

        Args:
            message: A MessageReceived object containing the received message.

        Returns:
            None. The message should be completely handled within this
            function.

        '''
        print('Handling received message')

    async def message_delivered(self, message_delivered):
        '''Handles all 'message delivered' events sent to the bot.

        Args:
            message_delivered: A MessageDelivered object containing the
                               received message.

        Returns:
            None. The message should be completely handled within this
            function.

        '''
        print('Handling delivered message')

    async def message_read(self, message_read):
        '''Handles all 'message read' events sent to the bot.

        Args:
            message_read: A MessageRead object containing the
                          received message.

        Returns:
            None. The message should be completely handled within this
            function.

        '''
        print('Handling read message')

    async def postback(self, postback):
        '''Handles all 'postback' events sent to the bot.

        Args:
            postback: A Postback object containing the
                      received message.

        Returns:
            None. The message should be completely handled within this
            function.

        '''
        print('Handling postback')

    async def referral(self, referral):
        '''Handles all 'referral' events sent to the bot.

        Args:
            referral: A Referral object containing the
                      received message.

        Returns:
            None. The message should be completely handled within this
            function.

        '''
        print('Handling referral')

    async def opt_in(self, opt_in):
        '''Handles all 'opt in' events sent to the bot.

        Args:
            opt_in: An OptIn object containing the
                    received message.

        Returns:
            None. The message should be completely handled within this
            function.

        '''
        print('Handling opt in')

    async def account_link(self, account_link):
        '''Handles all 'account linking' events sent to the bot.

        Args:
            opt_in: An AccountLink object containing the
                    received message.

        Returns:
            None. The message should be completely handled within this
            function.

        '''
        print('Handling account linking')



import abc

from .exceptions import BoomerangException


class Message():
    '''A class for messages to be sent to the Send API.

    Attributes:
        text: A string representing the text of the message.
        attachment: An Attachment object.
        quick_replies: A list of QuickReply objects.
        metadata: A string set by the developer that will be returned
                  to the webhook.
    '''

    def __init__(self, text=None, attachment=None, quick_replies=[],
                 metadata=None):

        # Ensure that either a text message or attachment is present.
        if text is None and attachment is None:
            error = 'Message objects require either text or an attachment'
            raise BoomerangException(error)

        self.text = text
        self.attachment = attachment
        self.quick_replies = quick_replies
        self.metadata = metadata

    def to_json(self):
        '''Converts the message to JSON.

        Returns:
            A dictionary holding the JSON representation of a message, which
            can be passed to the send() function.

        '''
        json = {}

        if self.text is not None:
            json['text'] = self.text

        if self.attachment is not None:
            json['attachment'] = self.attachment.to_json()

        if len(self.quick_replies) > 0:
            json['quick_replies'] = [quick_reply.to_json() for quick_reply in
                                     self.quick_replies]

        return json


class MediaAttachment:
    '''A class holding media attachments.

    Attributes:
        media_type: The type of the attachment. Can be one of
                    'image', 'audio', 'video', or 'file'.
        url: The url of the media as a string.

    '''

    def __init__(self, media_type, url):
        self.media_type = media_type
        self.url = url

    @classmethod
    def from_json(cls, json):
        '''Builds a MediaAttachment object from JSON provided by the API.

        Args:
            json: A dict containing the JSON representation that is converted
                  into the MediaAttachment object.

        Returns:
            A MediaAttachment object.

        '''
        return cls(json['type'], json['payload']['url'])

    def to_json(self):
        '''Converts the attachment to JSON.

        Returns:
            A dictionary holding the JSON representation of a media
            attachment.

        '''

        json = {'type': self.media_type,
                'payload': {'url': self.url}}

        return json


class LocationAttachment:
    '''A class holding location attachments.

    Attributes:
        latitude: A float of the latitude of the location.
        longitude: A float of the longitude of the location.

    '''

    def __init__(self, latitude, longitude):
        self.latitude = latitude
        self.longitude = longitude

    @classmethod
    def from_json(cls, json):
        '''Builds a LocationAttachment object from JSON provided by the API.

        Args:
            json: A dict containing the JSON representation that is converted
                  into the LocationAttachment object.

        Returns:
            A LocationAttachment object.

        '''
        return cls(json['payload']['coordinates']['lat'],
                   json['payload']['coordinates']['long'])


class Template(metaclass=abc.ABCMeta):
    '''The abstract base class for all template types.'''

    @abc.abstractmethod
    def to_json(self, payload):
        '''Implements the to_json() method for all template types.

        This method provides the JSON representation common to all templates,
        namely the 'type' and 'payload' attributes. Each subclass must call
        this method using super, and provide its own JSON representation of the
        payload.

        Args:
            payload: A dictionary JSON representation of the template's payload.

        Returns:
            A dictionary containing the complete JSON representation of
            the template.
        '''
        return {'type': 'template',
                'payload': payload}


class ButtonTemplate(Template):
    '''A class representing a button template.

    Attributes:
        text: The text of the message.
        buttons: A list of Button objects to include.

    '''
    def __init__(self, text, buttons):
        self.text = text
        self.buttons = buttons

    def to_json(self):
        '''Converts the template to JSON.

        Returns:
            A dictionary holding the JSON representation of the template.

        '''
        payload = {'template_type': 'button',
                   'text': self.text,
                   'buttons': [button.to_json() for button in self.buttons]}

        # Call the parent Template class with the payload
        return super().to_json(payload)


class ListTemplate(Template):
    '''A class representing a list template.

    Attributes:
        elements: A list of Elements to include in the list.
        button: A Button to include at the end of the list.
        compact: A boolean. If False, the first Element is expanded to fill
                 the top of the list. Otherwise, all items are rendered the
                 same way. Defaults to True.

    '''
    def __init__(self, elements, button=None, compact=True):

        # The Send API allows two to four elements only.
        if len(elements) in range(2, 5):
            self.elements = elements
        else:
            error = 'List templates require 2, 3, or 4 Elements'
            raise BoomerangException(error)

        self.button = button

        # Ensure the template is valid when in non-compact form.
        if not compact and elements[0].image_url is None:
            error = ('"image_url" must be present in the first element'
                     'of a non-compact list template')
            raise BoomerangException(error)

        self.compact = True

    def to_json(self):
        '''Converts the template to JSON.

        Returns:
            A dictionary holding the JSON representation of the template.

        '''
        payload = {'template_type': 'list',
                   'elements': [element.to_json() for element in
                                self.elements]}

        # While the Send API says that only one button may be used, it must
        # be within a list.
        if self.button is not None:
            payload['buttons'] = [self.button.to_json()]

        # Set the template style
        compact_styles = {True: 'compact', False: 'large'}
        payload['top_element_style'] = compact_styles[self.compact]

        # Call the parent Template class with the payload
        return super().to_json(payload)


class GenericTemplate(Template):
    '''A class representing a generic template.

    Attributes:
        elements: A list of Elements to include in the list.

    '''
    def __init__(self, elements):

        # The Send API allows up to 10 elements only.
        if len(elements) in range(1, 11):
            self.elements = elements
        else:
            error = 'Generic templates require between 1 and 10 Elements'
            raise BoomerangException(error)

    def to_json(self):
        '''Converts the template to JSON.

        Returns:
            A dictionary holding the JSON representation of the template.

        '''
        payload = {'template_type': 'generic',
                   'elements': [element.to_json() for element in
                                self.elements]}

        # Call the parent Template class with the payload
        return super().to_json(payload)


class Element:
    '''A class representing an element of the ListTemplate and GenericTemplate.

    Attributes:
        title: The name of the element.
        sub_title: A subtitle to display on the element.
        image_url: The URL of an image to display on the element.
        default_action: A DefaultAction to trigger when the element is pressed.
        buttons: A list of Buttons to add to the element.

    '''
    def __init__(self, title, sub_title=None, image_url=None,
                 default_action=None, buttons=None):
        self.title = title
        self.sub_title = sub_title
        self.image_url = image_url
        self.default_action = default_action
        self.buttons = buttons

    def to_json(self):
        '''Converts the element to JSON.

        Returns:
            A dictionary holding the JSON representation of the button.

        '''
        json = {'title': self.title}

        if self.sub_title is not None:
            json['subtitle'] = self.sub_title

        if self.image_url is not None:
            json['image_url'] = self.image_url

        if self.default_action is not None:
            json['default_action'] = self.default_action.to_json()

        if self.buttons is not None:
            json['buttons'] = [button.to_json() for button in self.buttons]

        return json


class DefaultAction:
    '''A class representing an Element's default action, i.e. launching a
    URL.

    Attributes:
        url: The url to open when the element is pressed.

    '''
    def __init__(self, url):
        self.url = url

    def to_json(self):
        '''Converts the action to JSON.

        Returns:
            A dictionary holding the JSON representation of the button.

        '''
        return {'type': 'web_url',
                'url': self.url}



"""
test_messages
----------------------------------

Tests for classes in the messages module.
"""

import pytest

from boomerang import messages, exceptions, buttons


def test_invalid_message():
    '''Ensures that Message objects require either a text message or Attachment
    when being initialised.'''
    with pytest.raises(exceptions.BoomerangException):
        messages.Message()


def test_text_message_json():
    '''Tests the to_json() functionality of the Message class, when it contains
    a text message.'''
    message = messages.Message(text='dummy_text')

    required_json = {'text': 'dummy_text'}

    assert message.to_json() == required_json


def test_media_attachment_json():
    '''Tests the to_json() functionality of the MediaAttachment class.'''

    attachment = messages.MediaAttachment('image', 'http://www.google.com')
    attachment_json = {'type': 'image',
                       'payload': {'url': 'http://www.google.com'}}
    message = messages.Message(attachment=attachment)

    assert attachment.to_json() == attachment_json
    assert message.to_json() == {'attachment': attachment_json}


def test_button_text_too_long():
    '''Ensures that Buttons raise an error when their text is too long.'''

    with pytest.raises(exceptions.BoomerangException):
        buttons.URLButton('This text is greater than 20',
                                       'http://www.google.com')


def test_button_template_json():
    '''Tests the to_json() functionality of the ButtonTemplate class.'''

    url_button = buttons.URLButton('Google', 'http://www.google.com')
    postback_button = buttons.PostbackButton('Postback', 'dummy_payload')
    call_button = buttons.CallButton('Call', '+15105551234')
    attachment = messages.ButtonTemplate('Select a URL', [url_button,
                                                          postback_button,
                                                          call_button])
    message = messages.Message(attachment=attachment)

    payload_json = {'template_type': 'button',
                    'text': 'Select a URL',
                    'buttons': [{'type': 'web_url',
                                 'url': 'http://www.google.com',
                                 'title': 'Google'},
                                {'type': 'postback',
                                 'payload': 'dummy_payload',
                                 'title': 'Postback'},
                                {'type': 'phone_number',
                                 'payload': '+15105551234',
                                 'title': 'Call'}]}

    message_json = {'attachment': {'type': 'template',
                                   'payload': payload_json}}

    assert message.to_json() == message_json


def test_button_template_json_share():
    '''Tests the to_json() functionality of the ButtonTemplate class,
    when a ShareButton is present.'''

    share_button = buttons.ShareButton()
    attachment = messages.ButtonTemplate('Select a URL', [share_button])
    message = messages.Message(attachment=attachment)

    payload_json = {'template_type': 'button',
                    'text': 'Select a URL',
                    'buttons': [{'type': 'element_share'}]}

    message_json = {'attachment': {'type': 'template',
                                   'payload': payload_json}}

    assert message.to_json() == message_json


def test_quick_reply():
    '''Tests the initialisation of the QuickReply class.'''

    # Ensure the QuickReply button_type can only be 'text' or 'location'
    with pytest.raises(exceptions.BoomerangException):
        buttons.QuickReply('image')

    # Ensure error is raised on missing text arguments
    with pytest.raises(exceptions.BoomerangException):
        buttons.QuickReply('text', text='dummy_text')

    with pytest.raises(exceptions.BoomerangException):
        buttons.QuickReply('text', payload='dummy_payload')

    # Ensure no other attributes are given when creating a location
    # QuickReply
    with pytest.raises(exceptions.BoomerangException):
        buttons.QuickReply('location', text='dummy_text')


def test_quick_reply_json():
    '''Tests the to_json() functionality of the QuickReply class.'''

    text_button = buttons.QuickReply('text', text='Button',
                                     payload='button_pressed',
                                     image_url='http://www.google.com')
    location_button = buttons.QuickReply('location')

    message = messages.Message(text='Choose a button',
                               quick_replies=[text_button,
                                              location_button])

    message_json = {'text': 'Choose a button',
                    'quick_replies': [{'content_type': 'text',
                                       'title': 'Button',
                                       'payload': 'button_pressed',
                                       'image_url': 'http://www.google.com'},
                                      {'content_type': 'location'}]}

    assert message.to_json() == message_json


def test_invalid_list_template():
    '''Tests the error-checking in the ListTemplate class.'''
    url_button = buttons.URLButton('Google', 'http://www.google.com')
    default_action = messages.DefaultAction('https://i.imgur.com/MBUyt0n.png')
    element = messages.Element('Element', sub_title='Element subtitle',
                               image_url='https://i.imgur.com/MBUyt0n.png',
                               default_action=default_action,
                               buttons=[url_button])

    # Between 2 and 4 elements must be provided
    with pytest.raises(exceptions.BoomerangException):
        list_template = messages.ListTemplate([])

    with pytest.raises(exceptions.BoomerangException):
        list_template = messages.ListTemplate([element, element, element,
                                               element, element])

    # Ensure that the first element of a ListTemplate in non-compact form
    # contains an image
    imageless_element = messages.Element('Element',
                                         default_action=default_action,
                                         buttons=[url_button])

    with pytest.raises(exceptions.BoomerangException):
        messages.ListTemplate([imageless_element, element],
                              compact=False)


def test_list_template_json():
    '''Tests the to_json() functionality of the ListTemplate class.'''

    url_button = buttons.URLButton('Google', 'http://www.google.com')
    default_action = messages.DefaultAction('https://i.imgur.com/MBUyt0n.png')
    element = messages.Element('Element', sub_title='Element subtitle',
                               image_url='https://i.imgur.com/MBUyt0n.png',
                               default_action=default_action,
                               buttons=[url_button])

    attachment = messages.ListTemplate([element, element],
                                       button=url_button)

    message = messages.Message(attachment=attachment)

    element_json = {'title': 'Element',
                    'subtitle': 'Element subtitle',
                    'image_url': 'https://i.imgur.com/MBUyt0n.png',
                    'default_action': {'type': 'web_url',
                                       'url': 'https://i.imgur.com/MBUyt0n.png'},
                    'buttons': [{'type': 'web_url',
                                 'title': 'Google',
                                 'url': 'http://www.google.com'}]}

    payload_json = {'template_type': 'list',
                    'elements': [element_json, element_json],
                    'buttons': [{'type': 'web_url',
                                 'url': 'http://www.google.com',
                                 'title': 'Google'}],
                    'top_element_style': 'compact'}

    message_json = {'attachment': {'type': 'template',
                                   'payload': payload_json}}

    assert message.to_json() == message_json


def test_generic_template_element_limit():
    '''Ensures that GenericTemplate raises an exception when the number
    of elements is greater than 10.'''
    default_action = messages.DefaultAction('https://i.imgur.com/MBUyt0n.png')

    element = messages.Element('Element', sub_title='Element subtitle',
                               image_url='https://i.imgur.com/MBUyt0n.png',
                               default_action=default_action)

    with pytest.raises(exceptions.BoomerangException):
        messages.GenericTemplate([element] * 11)


def test_generic_template_json():
    '''Tests the to_json() functionality of the GenericTemplate class.'''

    url_button = buttons.URLButton('Google', 'http://www.google.com')
    postback_button = buttons.PostbackButton('Postback', 'dummy_payload')
    default_action = messages.DefaultAction('https://i.imgur.com/MBUyt0n.png')

    element = messages.Element('Element', sub_title='Element subtitle',
                               image_url='https://i.imgur.com/MBUyt0n.png',
                               default_action=default_action,
                               buttons=[url_button, postback_button])

    attachment = messages.GenericTemplate([element, element])

    message = messages.Message(attachment=attachment)

    element_json = {'title': 'Element',
                    'subtitle': 'Element subtitle',
                    'image_url': 'https://i.imgur.com/MBUyt0n.png',
                    'default_action': {'type': 'web_url',
                                       'url': 'https://i.imgur.com/MBUyt0n.png'},
                    'buttons': [{'type': 'web_url',
                                 'title': 'Google',
                                 'url': 'http://www.google.com'},
                                {'type': 'postback',
                                 'payload': 'dummy_payload',
                                 'title': 'Postback'}]}

    payload_json = {'template_type': 'generic',
                    'elements': [element_json, element_json]}

    message_json = {'attachment': {'type': 'template',
                                   'payload': payload_json}}

    assert message.to_json() == message_json
